# -*- coding: utf-8 -*-
"""Functions related to definition of coherence models
"""
import itertools

import numpy as np


def get_coh_mat(freq, spat_df, coh_model='iec', dtype=np.float64, **kwargs):
    """Create coherence matrix for given frequencies and coherence model
    """
    if coh_model == 'iec':  # IEC coherence model
        coh_mat = get_iec_coh_mat(freq, spat_df, dtype=dtype, coh_model=coh_model,
                                  **kwargs)
    elif coh_model == 'iec3d':
        coh_mat = get_iec3d_coh_mat(freq, spat_df, dtype=dtype, coh_model=coh_model,
                                    **kwargs)
    else:  # unknown coherence model
        raise ValueError(f'Coherence model "{coh_model}" not recognized!')

    return coh_mat


def chunker(iterable, nPerChunks):
    """Return list of nPerChunks elements of an iterable"""
    it = iter(iterable)
    while True:
       chunk = list(itertools.islice(it, nPerChunks))
       if not chunk:
           return
       yield chunk


def get_iec_coh_mat(freq, spat_df, dtype=np.float64, ed=4, **kwargs):
    """Create IEC 61400-1 Ed. 3/4 coherence matrix for given frequencies
    """
    # preliminaries
    if ed not in [3, 4]:  # only allow edition 3
        raise ValueError('Only editions 3 or 4 are permitted.')
    if any([k not in kwargs for k in ['u_ref', 'l_c']]):  # check kwargs
        raise ValueError('Missing keyword arguments for IEC coherence model')
    freq = np.array(freq).reshape(-1, 1)  # nf x 1
    nf, ns = freq.size, spat_df.shape[1]
    # intermediate variables
    yz = spat_df.loc[['y', 'z']].values.astype(float)
    coh_mat = np.repeat(np.eye((ns), dtype=dtype)[None, :, :], nf, axis=0) # TODO use sparse matrix 
    exp_constant = np.sqrt( (1/ kwargs['u_ref'] * freq)**2 + (0.12 / kwargs['l_c'])**2).astype(dtype)  # nf x 1
    i_comp = np.arange(ns)[spat_df.iloc[0, :].values == 0]  # Selecting only u-components
    # loop through number of combinations, nPerChunks at a time to reduce memory impact
    for ii_jj in chunker(itertools.combinations(i_comp, 2), nPerChunks=10000):
        # get indices of point-pairs
        ii = np.array([tup[0] for tup in ii_jj])
        jj = np.array([tup[1] for tup in ii_jj])
        # calculate distances and coherences
        r = np.sqrt((yz[0, ii] - yz[0, jj])**2 + (yz[1, ii] - yz[1, jj])**2)  # nchunk
        coh_values = np.exp(-12 * np.abs(r) * exp_constant)  # nf x nchunk
        # Previous method (same math, different numerics)
        coh_mat[:, ii, jj] = coh_values  # nf x nchunk
        coh_mat[:, jj, ii] = np.conj(coh_values)
    return coh_mat


def get_iec3d_coh_mat(freq, spat_df, dtype=np.float64, **kwargs):
    """Create IEC-flavor coherence but for all 3 components
    """
    # preliminaries
    if any([k not in kwargs for k in ['u_ref', 'l_c']]):  # check kwargs
        raise ValueError('Missing keyword arguments for IEC coherence model')
    freq = np.array(freq).reshape(-1, 1)  # nf x 1
    nf, ns = freq.size, spat_df.shape[1]
    # intermediate variables
    yz = spat_df.loc[['y', 'z']].values.astype(float)
    coh_mat = np.repeat(np.eye((ns), dtype=dtype)[None, :, :], nf, axis=0) # TODO use sparse matrix 
    for ic in range(3):
        Lc = kwargs['l_c'] * [1, 2.7/8.1 , 0.66/8.1 ][ic]
        exp_constant = np.sqrt( (1/ kwargs['u_ref'] * freq)**2 + (0.12 / Lc)**2).astype(dtype)
        i_comp = np.arange(ns)[spat_df.iloc[0, :].values == ic]  # Selecting only u-components
        # loop through number of combinations, nPerChunks at a time to reduce memory impact
        for ii_jj in chunker(itertools.combinations(i_comp, 2), nPerChunks=10000):
            # get indices of point-pairs
            ii = np.array([tup[0] for tup in ii_jj])
            jj = np.array([tup[1] for tup in ii_jj])
            # calculate distances and coherences
            r = np.sqrt((yz[0, ii] - yz[0, jj])**2 + (yz[1, ii] - yz[1, jj])**2)
            coh_values = np.exp(-12 * np.abs(r) * exp_constant)
            # Previous method (same math, different numerics)
            coh_mat[:, ii, jj] = coh_values  # nf x nchunk
            coh_mat[:, jj, ii] = np.conj(coh_values)
    return coh_mat



# -*- coding: utf-8 -*-
"""Define how the turbulence standard deviation varies with k, y and z.

You can either use the built-in models (see below) or create your own function
to model the spatial variation of the turbulence standard deviation.
"""
import warnings

import numpy as np

from pyconturb._utils import _DEF_KWARGS, interpolator


def get_sig_values(spat_df, sig_func, **kwargs):
    """Turbulence standard deviation for points/components in ``spat_df``.

    The ``sig_func`` must be a function of the form::

        sig_values = sig_func(spat_df, **kwargs)

    where k, y and z can be floats, np.arrays or pandas.Series. You can use
    the functions built into PyConTurb (see below) or define your own custom
    function. The output is assumed to be in m/s.

    Parameters
    ----------
    spat_df : pandas.DataFrame
        Spatial information on the points to simulate. Must have columns
        ``[k, p_id, x, y, z]``, and each of the ``n_sp`` rows corresponds
        to a different spatial location and turbuine component (u, v or
        w).
    sig_func : function
        Function to map k, y and z to the turbulence standard deviation in m/s.
    **kwargs
        Keyword arguments to pass into ``sig_func``.

    Returns
    -------
    sig_values : np.array
        [m/s] Turbulence standard deviation(s) for the given spatial locations(s)
        /component(s). Dimension is ``(n_sp,)``.
    """
    return sig_func(spat_df, **kwargs)


def data_sig(spat_df, con_tc=None, warn_datacon=True, **kwargs):
    """Turbulence standard deviation interpolated from a TimeConstraint object.

    See the Examples and/or Reference Guide for details on the interpolator logic or for
    how to construct a TimeConstraint object. Note that this function uses the
    biased estimator for the standard deviation (i.e., NumPy's default ``np.std``).

    Note! If a component is requested for which there is no constraint, then this
    function will  try to use iec_sig instead. Use the `warn_datacon` option to 
    disable the warning about this.

    Parameters
    ----------
    spat_df : pandas.DataFrame
        Spatial information on the points to simulate. Must have columns
        ``[k, p_id, x, y, z]``, and each of the ``n_sp`` rows corresponds
        to a different spatial location and turbuine component (u, v or
        w).
    con_tc : pyconturb.TimeConstraint
        [-] Constraint object. Must have correct format; see documentation on
        PyConTurb's TimeConstraint object for more details.
    warn_datacon : boolean
        [-] Warn if a requested component does not have a constraint, which results in
        an attempt at using IEC values. Default is True.
    **kwargs
        Unused (optional) keyword arguments.

    Returns
    -------
    sig_values : np.array
        [m/s] Turbulence standard deviation(s) at the specified location(s). Dimension
        is ``(n_sp,)``.
    """
    if con_tc is None:
        raise ValueError('No data provided!')
    k, y, z = spat_df.loc[['k', 'y', 'z']].values
    out_array = np.empty_like(y, dtype=float)
    for kval in np.unique(k):  # loop over passed-in components
        out_mask = np.isclose(k, kval)
        # make sure that kval is in con_tc
        if not np.any(np.isclose(kval, con_tc.loc['k'])):
            if warn_datacon:  # throw warning if requested
                warnings.warn(f'Requested component "{kval}" does not exist in constraints! '
                              + 'Cannot interpolate sig. Setting to zero.',
                              Warning, stacklevel=2)
            out_array[out_mask] = iec_sig(spat_df.iloc[:, out_mask], **kwargs)
            continue
        con_mask = (con_tc.loc['k'] == kval).values
        ypts = con_tc.iloc[2, con_mask].values.astype(float)
        zpts = con_tc.iloc[3, con_mask].values.astype(float)
        vals = np.std(con_tc.iloc[4:, con_mask], axis=0).astype(float)
        out_array[out_mask] = interpolator((ypts, zpts), vals, (y[out_mask], z[out_mask]))
    return out_array


def iec_sig(spat_df, turb_class=_DEF_KWARGS['turb_class'], **kwargs):
    """Turbulence standard deviation as specified in IEC 61400-1 Ed. 3.

    Parameters
    ----------
    spat_df : pandas.DataFrame
        Spatial information on the points to simulate. Must have columns
        ``[k, p_id, x, y, z]``, and each of the ``n_sp`` rows corresponds
        to a different spatial location and turbuine component (u, v or
        w).
    z : array-like
        [m] Location of point(s) in the vertical direction. Can be int/float,
        np.array or pandas.Series.
    turb_class : str, optional
        [-] Turbulence class.
    **kwargs
        Unused (optional) keyword arguments.

    Returns
    -------
    sig_values : np.array
        [m/s] Turbulence standard deviation(s) at the specified location(s). Dimension
        is ``(n_sp,)``.
    """
    kwargs = {**{'turb_class': turb_class}, **kwargs}  # add dflts if not given
    assert kwargs['turb_class'].lower() in 'abc', 'Invalid or no turbulence class!'
    i_ref = {'a': 0.16, 'b': 0.14, 'c': 0.12}[kwargs['turb_class'].lower()]
    k = spat_df.loc[['k']].values.astype(int)
    y, z = spat_df.loc[['y', 'z']].values
    sig1 = i_ref * (0.75 * kwargs['u_ref'] + 5.6)  # std dev in u
    sig_k = sig1 * np.asarray(1.0 * (k == 0) + 0.8 * (k == 1) + 0.5 * (k == 2))
    return np.array(sig_k, dtype=float).squeeze()



# -*- coding: utf-8 -*-
"""Functions related to definitions of spectral models
"""
import numpy as np

from pyconturb.spectral_models import get_spec_values
from pyconturb.sig_models import get_sig_values
from pyconturb._utils import get_freq


def get_magnitudes(spat_df, spec_func, sig_func, **kwargs):
    """"""
    t, freq = get_freq(**kwargs)
    spc_arr = get_spec_values(freq, spat_df, spec_func, **kwargs)
    mags_arr = spc_to_mag(spat_df, spc_arr, sig_func, **kwargs)
    return mags_arr


def spc_to_mag(spat_df, spc_arr, sig_func, **kwargs):
    """Convert spectral array to magnitudes (sets DC component to 0)
    """
    # get unscaled magnitudes
    t, freq = get_freq(**kwargs)
    n_t, df = t.size, freq[1]
    mags = np.sqrt(spc_arr * df / 2)  # (nf, nsp)
    mags[0, :] = 0.  # mean is zero to make math easier
    # scale to get the correct ti
    std_theo = get_sig_values(spat_df, sig_func, **kwargs)  # (n_sp,)
    if n_t == 2:
        std_now = np.sqrt(n_t/(n_t-1) * np.sum(np.abs(mags)**2, axis=0))
    else:
        std_now = np.sqrt(n_t/(n_t-1)
                          * (2*np.sum(np.abs(mags)**2, axis=0)
                             - ((n_t+1) % 2)*np.abs(mags[-1, :])**2))
    alpha = std_theo / std_now
    return (alpha * mags).astype(float)  # (nf, nsp)



# -*- coding: utf-8 -*-
"""Define how the turbulence power spectrum varies with k, y and z.

You can either use the built-in models (see below) or create your own function.
Note that these spectra are continuous, one-sided spectra. The conversion from
continuous to discrete spectra and scaling to the appropriate variance is done
during simulation.
"""
import warnings

import numpy as np

from pyconturb._utils import _DEF_KWARGS, interpolator


def get_spec_values(f, spat_df, spec_func, **kwargs):
    """Power spectral density (PSD) for points/components in ``spat_df``.

    The ``spec_func`` must be a function of the form::

        spec_values = spec_func(f, spat_df, **kwargs).

    You can use the functions built into PyConTurb (see below) or define your own custom
    function. The output is assumed to be in (m^2/s^2)/Hz = m^2/s. There is no
    need to scale to the correct variance -- the spectrum is scaled during simulation
    in order to produce the standard deviation specified by ``sig_func``.

    Parameters
    ----------
    f : array-like
        [Hz] Frequency(s) for which PSD is to be calculated. Size is ``n_f``.
    spat_df : pandas.DataFrame
        Spatial information on the points to simulate. Must have columns
        ``[k, x, y, z]``, and each of the ``n_sp`` rows corresponds
        to a different spatial location and turbuine component (u, v or
        w).
    spec_func : function
        Function to map k, y and z to the continuous, one-sided power spectral
        density in m^2/s.
    **kwargs
        Keyword arguments to pass into ``spec_func``.

    Returns
    -------
    spec_values : np.array
        [m^2/s] PSD values for the given spatial locations(s)/component(s).
        Dimension is ``(n_f, n_sp,)``.
    """
    return spec_func(f, spat_df, **kwargs)


def data_spectrum(f, spat_df, con_tc=None, warn_datacon=True, **kwargs):
    """Power spectrum interpolated from a TimeConstraint object.

    See the Examples and/or Reference Guide for details on the interpolator logic or for
    how to construct a TimeConstraint object. Note that this function can only return
    frequency values as defined by T, dt in ``con_tc``.

    Note! If a component is requested for which there is no constraint, then this
    function will try to use the IEC spectra instead. Use the `warn_datacon` option to 
    disable the warning about this.

    Parameters
    ----------
    f : array-like
        [Hz] Frequency(s) for which PSD is to be calculated. Size is ``n_f``.
    spat_df : pandas.DataFrame
        Spatial information on the points to simulate. Must have columns
        ``[k, x, y, z]``, and each of the ``n_sp`` rows corresponds
        to a different spatial location and turbuine component (u, v or
        w).
    con_tc : pyconturb.TimeConstraint
        [-] Constraint object. Must have correct format; see documentation on
        PyConTurb's TimeConstraint object for more details.
    warn_datacon : boolean
        [-] Warn if a requested component does not have a constraint, which results in
        an attempt at using the Kaimal spectrum. Default is True.
    **kwargs
        Unused (optional) keyword arguments.

    Returns
    -------
    spec_values : np.array
        [m^2/s] PSD values for the given spatial locations(s)/component(s).
        Dimension is ``(n_f, n_sp,)``.
    """
    if con_tc is None:
        raise ValueError('No data provided!')
    # get time array and isolate/sanitize useful variables
    time_df = con_tc.get_time()
    df, nt = 1 / (time_df.index[-1] + time_df.index[1]), time_df.shape[0]
    f_reqs = np.array(f, ndmin=1)
    k, y, z = spat_df.loc[['k', 'y', 'z']].values
    # initialize spectral values
    spec_values = np.empty((f_reqs.size, y.size), dtype=float)
    for kval in np.unique(k):  # loop over the unique requested k values
        spec_mask = np.isclose(k, kval)  # cols in spec_values corr. to this kval
        # make sure that kval is in con_tc
        if not np.any(np.isclose(kval, con_tc.loc['k'])):
            if warn_datacon:  # throw warning if requested
                warnings.warn(f'Requested component "{kval}" does not exist in constraints! '
                              + 'Cannot interpolate spectra. Trying to use Kaimal instead.',
                              Warning, stacklevel=2)
            spec_values[:, spec_mask] = kaimal_spectrum(f, spat_df.iloc[:, spec_mask], **kwargs)
            continue
        con_mask = (con_tc.loc['k'] == kval).values
        ypts = con_tc.iloc[2, con_mask].values.astype(float)
        zpts = con_tc.iloc[3, con_mask].values.astype(float)
        k_time = con_tc.iloc[4:, con_mask]
        spc_data = 2 * np.abs(np.fft.rfft(k_time, axis=0) / nt)**2 / df
        freq = np.arange(spc_data.shape[0]) * df
        for i_f, f_req in enumerate(f_reqs):
            f_idx = np.argmax(np.isclose(f_req, freq))  # idx of req'd f in con_tc arr
            vals = spc_data[f_idx, :]  # pull out that frequency/component
            spec_values[i_f, spec_mask] = interpolator((ypts, zpts), vals,
                                                       (y[spec_mask], z[spec_mask]))
    return spec_values


def kaimal_spectrum(f, spat_df, u_ref=_DEF_KWARGS['u_ref'], **kwargs):
    """Kaimal PSD as specified in IEC 61400-1 Ed. 3.
    f is (nf,); k, y and z are (n_sp,), u_ref is float or int. returns (nf, n_sp,).
    No std scaling -- that's done with the magnitudes.

    Parameters
    ----------
    f : array-like
        [Hz] Frequency(s) for which PSD is to be calculated. Size is ``n_f``.
    spat_df : pandas.DataFrame
        Spatial information on the points to simulate. Must have columns
        ``[k, x, y, z]``, and each of the ``n_sp`` rows corresponds
        to a different spatial location and turbuine component (u, v or
        w).
    u_ref : int/float, optional
        [m/s] Mean wind speed at reference height.
    **kwargs
        Unused (optional) keyword arguments.

    Returns
    -------
    spec_values : np.array
        [m^2/s] PSD values for the given spatial locations(s)/component(s).
        Dimension is ``(n_f, n_sp,)``.
    """
    kwargs = {**{'u_ref': u_ref}, **kwargs}  # add dflts if not given
    k, y, z = spat_df.loc[['k', 'y', 'z']].values  # in case pd.series passed in
    f = np.reshape(f, (-1, 1))  # convert to column array
    lambda_1 = 0.7 * z * (z < 60) + 42 * (z >= 60)  # length scale changes with z
    l_k = lambda_1 * (8.1 * np.isclose(k, 0) + 2.7 * np.isclose(k, 1) + 0.66 * np.isclose(k, 2))
    tau = np.reshape((l_k / kwargs['u_ref']), (1, -1))  # L_k / U. row vector
    spec_values = (4 * tau) / np.power(1. + 6 * tau * f, 5. / 3.)  # Kaimal 1972
    return spec_values.astype(float)  # pandas causes object issues, ensure float



"""Reading HAWC2 results files (.set and .dat)

Notes
-----
Much of this was copied, modified, and/or cleaned from DTU's Wind Energy
Toolbox. https://gitlab.windenergy.dtu.dk/toolbox/WindEnergyToolbox
"""
import os

import numpy as np
import pandas as pd


# hard-code parameters from .set file format
_sel_lineno_info = 8  # line no. in sel with info
_sel_lineno_chanstart = 12  # line no. channel start
_sel_col_wds = [13, 44, 55]  # character widths in sel file
_sel_df_cols = ['channel', 'var_desc', 'units', 'notes', 'scale']


def sel_to_df(path, unique=True):
    """Load data from a HAWC2 sel file into a pandas dataframe

    Arguments
    ---------
    path : str
        Path to .sel file
    unique : boolean
        Whether to also create unique channel names

    Returns
    -------
    res_df : pandas.DataFrame
        Pandas dataframe with results from HAWC2
    """

    # load info from sel file
    path = os.path.splitext(path)[0]  # remove ext if given
    sel_path = f'{path}.sel'  # path to sel file
    with open(sel_path, 'r') as sel_fid:
        sel_lines = sel_fid.readlines()
    line_info = sel_lines[_sel_lineno_info].split()
    n_chnls = int(line_info[1])
    fmt = line_info[3]

    # load data from sel file
    sel_df = pd.DataFrame(np.empty((n_chnls, len(_sel_df_cols))),
                          columns=_sel_df_cols)  # dataframe with sel data
    sel_df['scale'] = 1.  # initialize scale factor to one
    for i_line in range(n_chnls):
        sel_line = sel_lines[i_line + _sel_lineno_chanstart]
        sel_df.iloc[i_line, 0] = int(sel_line[:_sel_col_wds[0]].strip())  # chn
        sel_df.iloc[i_line, 1] = sel_line[_sel_col_wds[0]:
                                          _sel_col_wds[1]].strip()  # varble
        sel_df.iloc[i_line, 2] = sel_line[_sel_col_wds[1]:
                                          _sel_col_wds[2]].strip()  # units
        sel_df.iloc[i_line, 3] = sel_line[_sel_col_wds[2]:].strip()  # descptn
        if fmt.lower() == 'binary':  # get scale if binary
            sel_df.iloc[i_line, -1] = float(sel_lines[i_line +
                                                      _sel_lineno_chanstart
                                                      + n_chnls + 2])

    if unique:
        sel_df['unique'] = get_unique_chnl_names(sel_df)

    return sel_df


def get_unique_chnl_names(sel):
    """Unique channel names from hawc2 sel file

    This assumes the user made use of the hashtag's ability to custom name
    their own channels.
    """
    if isinstance(sel, str):
        sel_df = sel_to_df(sel)
    elif isinstance(sel, pd.DataFrame):
        sel_df = sel.copy()
    else:
        raise ValueError(f'Unknown input type {type(sel)}')
    sel_df['unique'] = ''
    sel_df.loc[sel_df.var_desc == 'Time', 'unique'] = 'time'
    for idx in sel_df[sel_df['notes'].str.contains('Free wind speed Vx')].index:
        row = sel_df.loc[idx, :]
        sel_df.loc[idx, 'unique'] = row.notes.split()[-1].replace('wind_', 'vxg_')
    for idx in sel_df[sel_df['notes'].str.contains('Free wind speed Vy')].index:
        row = sel_df.loc[idx, :]
        sel_df.loc[idx, 'unique'] = row.notes.split()[-1].replace('wind_', 'vyg_')
    for idx in sel_df[sel_df['notes'].str.contains('Free wind speed Vz')].index:
        row = sel_df.loc[idx, :]
        sel_df.loc[idx, 'unique'] = row.notes.split()[-1].replace('wind_', 'vzg_')

    return sel_df.unique.values


def get_num_scans(path):
    """get number of scans from sel file
    """
    # load info from sel file
    path = os.path.splitext(path)[0]  # remove ext if given
    sel_path = f'{path}.sel'  # path to sel file
    with open(sel_path, 'r') as sel_fid:
        sel_lines = sel_fid.readlines()
    line_info = sel_lines[_sel_lineno_info].split()
    return int(line_info[0])


def dat_to_df(path, n_scns=None, channel_names=None, sel_df=None,
              unique=True):
    """HAWC2 binary .dat file to pandas dataframe
    """
    # load necessary info if not passed in
    if n_scns is None:  # get number of scans if not passed in
        n_scns = get_num_scans(path)
    if channel_names is None:  # get channel names if not passed in
        channel_names = get_unique_chnl_names(path)
    if sel_df is None:  # get .sel dataframe if not passed in
        sel_df = sel_to_df(path)

    # load data in dat file
    dat_path = os.path.splitext(path)[0] + '.dat'  # path to dat file
    with open(dat_path, 'rb') as dat_fid:
        dat_df = pd.DataFrame(np.zeros((n_scns, len(channel_names))),
                              columns=channel_names)
        for i in range(sel_df.shape[0]):
            dat_fid.seek(i * n_scns * 2, 0)
            dat_df.iloc[:, i] = np.fromfile(dat_fid, 'int16', n_scns) * \
                sel_df.loc[i, 'scale']

    if unique:
        dat_df.columns = get_unique_chnl_names(sel_df)
    return dat_df
